import logging
import pathlib
from typing import Optional

import typer

from qfmu import __version__, model
from qfmu.utils import build_fmu, str_to_arr, str_to_mat

app = typer.Typer(rich_markup_mode="rich")


@app.command()
def ss(
    A: Optional[str] = typer.Option(
        None, "--A", "-A", help="A matrix as a json str, zero if empty"
    ),  # noqa: E501
    B: Optional[str] = typer.Option(
        None, "--B", "-B", help="B matrix as a json str, inferred if empty"
    ),  # noqa: E501
    C: Optional[str] = typer.Option(
        None,
        "--C",
        "-C",
        help="C matrix as a json str, identity matrix if empty, unless D is provided",
    ),  # noqa: E501
    D: Optional[str] = typer.Option(
        None, "--D", "-D", help="D matrix as a json str, inferred if empty"
    ),  # noqa: E501
    x0: Optional[str] = typer.Option(
        None,
        "--x0",
        "-x0",
        help="Initial state vector as a json str. Zero vector if empty",
    ),
    u0: Optional[str] = typer.Option(
        None,
        "--u0",
        "-u0",
        help="Initial input vector as a json str. Zero vector if empty",
    ),
    dt: float = typer.Option(
        0.001, "--dt", "-h", help="Euler integrator step size", min=0.0
    ),
    output: pathlib.Path = typer.Option(
        "./q.fmu",
        "--output",
        "-o",
        help="FMU Output path",
        writable=True,
        file_okay=True,
        dir_okay=False,
        resolve_path=True,
    ),
):
    """
    Generate a continuous-time state space system fmu

    Examples:

    qfmu ss -A "[[1,2],[3,4]]" -o ./q.fmu
    - A[2, 2] = [[1.0, 2.0], [3.0, 4.0]]
    - B[2, 0] = [[], []]
    - C[2, 2] = [[1.0, 0.0], [0.0, 1.0]]
    - D[2, 0] = [[], []]
    """
    # Hanlde case where only D is provided
    A = str_to_mat(A) if A is not None else None
    B = str_to_mat(B) if B is not None else None
    C = str_to_mat(C) if C is not None else None
    D = str_to_mat(D) if D is not None else None
    x0 = str_to_arr(x0) if x0 is not None else None
    u0 = str_to_arr(u0) if u0 is not None else None

    # Get filename as identifier
    if output.suffix != ".fmu":
        raise ValueError("Output file must be an FMU")
    identifier = output.stem

    # Construct a state space model
    m = model.StateSpace(A, B, C, D, x0, u0)

    # Build FMU
    build_fmu(m, output, identifier=identifier, dt=dt)


@app.command()
def tf(
    num: str = typer.Option("--num", help="Numerator polynomial coefficients"),
    den: str = typer.Option("--den", help="Denominator polynomial coefficients"),
    x0: Optional[float] = typer.Option(
        None, "--x0", "-x0", help="Initial state vector. Zero vector if empty"
    ),
    u0: Optional[float] = typer.Option(
        None, "--u0", "-u0", help="Initial input vector. Zero vector if empty"
    ),
    dt: float = typer.Option(
        0.001, "--dt", "-h", help="Euler integrator step size", min=0.0
    ),
    output: pathlib.Path = typer.Option(
        "./q.fmu",
        "--output",
        "-o",
        help="FMU Output path",
        writable=True,
        file_okay=True,
        dir_okay=False,
        resolve_path=True,
    ),
):
    """Generate a continuous-time transfer function fmu

    tf(num, den) =

    num[0]*s**(n-1) + ... + num[n-1]*s + num[n]
    -----------------------------------
    den[0]*s**(n-1) + ... + den[n-1]*s + den[n]
    """
    # Get filename as identifier
    if output.suffix != ".fmu":
        raise ValueError("Output file must be an FMU")
    identifier = output.stem

    # Construct a state space model
    m = model.TransferFunction(str_to_arr(num), str_to_arr(den), x0, u0)

    # Build FMU
    build_fmu(m, output, identifier=identifier, dt=dt)


@app.command()
def pid(
    kp: float = typer.Option(0.0, "--kp", help="Proportional gain"),
    ki: float = typer.Option(0.0, "--ki", help="Integral gain"),
    kd: float = typer.Option(0.0, "--kd", help="Derivative gain"),
    T: float = typer.Option(
        0.0, "--T", "-T", help="First-order derivative filter time constant"
    ),  # noqa: E501
    x0: Optional[float] = typer.Option(
        None, "--x0", "-x0", help="Initial state vector. Zero vector if empty"
    ),
    u0: Optional[float] = typer.Option(
        None, "--u0", "-u0", help="Initial input vector. Zero vector if empty"
    ),
    dt: float = typer.Option(
        0.001, "--dt", "-h", help="Euler integrator step size", min=0.0
    ),
    output: pathlib.Path = typer.Option(
        "./q.fmu",
        "--output",
        "-o",
        help="FMU Output path",
        writable=True,
        file_okay=True,
        dir_okay=False,
        resolve_path=True,
    ),
):
    """Generate a PID controller fmu

    pid(kp, ki, kd) =

    kp + ki/s + kd*s/(tf*s + 1)

    """
    # Get filename as identifier
    if output.suffix != ".fmu":
        raise ValueError("Output file must be an FMU")
    identifier = output.stem

    # Construct a state space model
    m = model.PID(kp, ki, kd, T, x0, u0)

    # Build FMU
    build_fmu(m, output, identifier=identifier, dt=dt)


def version_callback(value: bool):
    if value:
        typer.echo(f"qfmu version: {__version__}")
        raise typer.Exit()


@app.callback()
def main(
    version: bool = typer.Option(
        None, "--version", callback=version_callback, is_eager=True
    ),
):
    logging.basicConfig(level=logging.INFO)
    return


if __name__ == "__main__":
    app()



import logging
import math
from typing import Optional

import numpy as np
from scipy import signal

from qfmu.model.lti import LTI


class PID(LTI):
    def __init__(
        self,
        kp: float = 0.0,
        ki: float = 0.0,
        kd: float = 0.0,
        T: float = 0.0,
        x0: Optional[float] = None,
        u0: Optional[float] = None,
    ) -> None:
        has_P = not math.isclose(kp, 0.0)
        has_I = not math.isclose(ki, 0.0)
        has_D = not math.isclose(kd, 0.0)

        if math.isclose(T, 0.0) and has_D:
            raise ValueError("T must be greater than zero")

        P = signal.TransferFunction([kp], [1.0]) if has_P else None  # noqa: E741
        I = signal.TransferFunction([ki], [1.0, 0.0]) if has_I else None  # noqa: E741
        D = (  # noqa: E741
            signal.TransferFunction([kd, 0.0], [T, 1.0]) if has_D else None
        )

        if sum([has_P, has_I, has_D]) == 0:
            raise ValueError("At least one of kp, ki, kd must be non-zero")
        elif sum([has_P, has_I, has_D]) == 1:
            self.m = (
                P.to_ss()
                if P is not None
                else I.to_ss()
                if I is not None
                else D.to_ss()
            )
        elif sum([has_P, has_I, has_D]) == 2:
            if has_P and has_I:
                self.m = signal.TransferFunction([kp, ki], [1.0, 0.0]).to_ss()
            elif has_P and has_D:
                self.m = signal.TransferFunction([kp * T + kd, kd], [T, 1.0]).to_ss()
            else:
                self.m = I.to_ss() + D.to_ss()
        else:
            self.m = signal.TransferFunction([kp, ki], [1.0, 0.0]).to_ss() + D.to_ss()

        super().__init__(
            self.m.A.shape[0],
            1,
            1,
            np.array([x0]) if x0 is not None else None,
            np.array([u0]) if u0 is not None else None,
        )

        logging.info(f"P = {P}")
        logging.info(f"I = {I}")
        logging.info(f"D = {D}")

    @property
    def A(self) -> np.ndarray:
        return self.m.A

    @property
    def B(self) -> np.ndarray:
        return self.m.B

    @property
    def C(self) -> np.ndarray:
        return self.m.C

    @property
    def D(self) -> np.ndarray:
        return self.m.D



import sys

import numpy as np
import pytest

from qfmu.utils import find_vcvarsall_location, str_to_arr, str_to_mat


def test_str_to_mat():
    assert np.array_equal(str_to_mat("[[1,2],[3,4]]"), np.array([[1, 2], [3, 4]]))


def test_str_to_arr():
    assert np.array_equal(str_to_arr("[1,2,3,4]"), np.array([1, 2, 3, 4]))


@pytest.mark.skipif(
    not sys.platform.startswith("win"), reason="This test is only for Windows"
)
def test_find_vcvarsall_location():
    assert find_vcvarsall_location() is not None



import datetime
import json
import logging
import os
import pathlib
import shutil
import subprocess
import tempfile
import time
import uuid

import numpy as np
from jinja2 import Environment, FileSystemLoader, select_autoescape

from qfmu import __include_path__, __platform__, __template_path__, __version__
from qfmu.codegen.utils import array2cstr
from qfmu.model.lti import LTI

env = Environment(
    loader=FileSystemLoader(__template_path__),
    autoescape=select_autoescape(),
    trim_blocks=True,
)
env.filters["array2cstr"] = array2cstr


def str_to_mat(data: str) -> np.ndarray:
    m = np.array(json.loads(data), dtype=float)
    if len(m.shape) != 2:
        raise ValueError("Invalid matrix format.")
    return m


def str_to_arr(data: str) -> np.ndarray:
    m = np.array(json.loads(data), dtype=float)
    if len(m.shape) != 1:
        raise ValueError("Invalid array format.")
    return m


def find_vcvarsall_location():
    try:
        # Find vswhere.exe
        vswhere = rf'{os.environ["ProgramFiles(x86)"]}\Microsoft Visual Studio\Installer\vswhere.exe'  # noqa: E501
        # Run vswhere and capture its output as JSON.
        vswhere_output = subprocess.check_output(
            [
                f"{vswhere}",
                "-latest",
                "-products",
                "*",
                "-requires",
                "Microsoft.Component.MSBuild",
                "-format",
                "json",
            ]
        )

        # Parse the JSON output.
        vswhere_data = json.loads(vswhere_output.decode("utf-8"))

        if vswhere_data and isinstance(vswhere_data, list):
            # Extract the installation path from the vswhere data.
            installation_path = vswhere_data[0].get("installationPath")

            if installation_path:
                # Construct the full path to vcvarsall.bat.
                vcvarsall_path = os.path.join(
                    installation_path, "VC", "Auxiliary", "Build", "vcvarsall.bat"
                )
                return vcvarsall_path
    except subprocess.CalledProcessError:
        pass

    return None


def compile_dll(src_dir: pathlib.Path, identifier: str) -> pathlib.Path:
    if __platform__.startswith("win"):
        compiler = "vc"
    elif __platform__.startswith("darwin"):
        compiler = "clang"
    else:
        compiler = "gcc"

    if compiler == "vc":
        target = identifier + ".dll"
        toolset = "x86_amd64" if "64" in __platform__ else "x86"
        compiler_options = "/Oy /Ob1 /Oi /LD"
        cmd = rf'call "{find_vcvarsall_location()}" {toolset}'
        cmd += f" && cl {compiler_options} /I./include /DDISABLE_PREFIX /Fe{target} shlwapi.lib fmi2model.c"  # noqa: E501
    elif compiler == "gcc":
        target = identifier + ".so"
        cmd = "gcc -c -I ./include -fPIC -DDISABLE_PREFIX fmi2model.c "
        cmd += f" && gcc -static-libgcc -shared -o{target} *.o -lm"
    elif "darwin" in __platform__:
        target = identifier + ".dylib"
        cmd = "clang -c -arch x86_64 -arch arm64 -I ./include -DDISABLE_PREFIX fmi2model.c"  # noqa: E501
        cmd += f" && clang -shared -arch x86_64 -arch arm64 -o{target} *.o -lm"
    else:
        raise RuntimeError("Unknown compiler")

    wd = os.getcwd()
    os.chdir(src_dir)
    status = os.system(cmd)
    os.chdir(wd)

    dll_path = src_dir / target
    if status != 0 or not dll_path.exists():
        raise Exception("Failed to compile shared library")

    return dll_path


def build_fmu(
    model: LTI,
    output: pathlib.Path = pathlib.Path("."),
    identifier: str = "model",
    dt: float = 0.001,
) -> None:
    _guid = str(uuid.uuid1())
    _datetime = datetime.datetime.now().strftime("%Y-%m-%dT%H:%M:%S.%f")[:-3]

    # TODO: Support fmi3 by loading different templates
    fmu_model_tmpl = env.get_template("fmi2model.jinja")
    fmu_desc_tmpl = env.get_template("fmi2modelDescription.jinja")

    # TODO: Fix me on Windows
    with tempfile.TemporaryDirectory() as tmpdir:
        tmpdir = pathlib.Path(tmpdir)
        # Create folder structure
        logging.debug(f"Creating folder structure at {tmpdir}")
        bin_dir = tmpdir / "binaries"
        bin_dir.mkdir()
        platform_dir = bin_dir / __platform__
        platform_dir.mkdir()
        src_dir = tmpdir / "sources"
        src_dir.mkdir()

        # Write source files
        logging.debug(f"Writing source files to {src_dir}")
        fmu_model = fmu_model_tmpl.render(
            model=model, identifier=identifier, version=__version__, guid=_guid, dt=dt
        )
        with open(src_dir / "fmi2model.c", "w") as f:
            f.write(fmu_model)
        # Copy header files to source folder
        shutil.copytree(__include_path__, src_dir / "include")

        # Compile model
        logging.debug("Compiling dll")
        dll_path = compile_dll(src_dir, identifier=identifier)
        shutil.move(dll_path, platform_dir / dll_path.name)

        # Write model description
        logging.debug("Writing model description")
        fmu_desc = fmu_desc_tmpl.render(
            model=model,
            identifier=identifier,
            version=__version__,
            guid=_guid,
            datetime=_datetime,
            dt=dt,
        )
        with open(tmpdir / "modelDescription.xml", "w") as f:
            f.write(fmu_desc)

        # Generate FMU
        logging.debug("Generating FMU")
        zippath = shutil.make_archive(identifier, "zip", tmpdir)
        shutil.move(zippath, str(output.parent / f"{identifier}.fmu"))

        logging.info(f"FMU generated successfully at {output}")

        # FIXME:
        # - vc holds on to files for a while and prevents deletion
        # - need to delete compiler generated temp files
        if __platform__.startswith("win"):
            time.sleep(1)  # thank you windose...
            # for file in src_dir.glob("*"):
            #     if file.suffix not in [".c", ".h"]:
            #         file_path = src_dir / file
            #         file_path.unlink()
            #         print(f"Removed file: {file_path}")

        shutil.rmtree(tmpdir)



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Optional

import numpy as np
import numpy.typing as npt
from annotated_types import Ge
from typing_extensions import Annotated


@dataclass
class VR0:
    x: int
    der: int
    x0: int
    u: int
    u0: int
    y: int


@dataclass
class VR:
    x: np.ndarray
    der: np.ndarray
    x0: np.ndarray
    u: np.ndarray
    u0: np.ndarray
    y: np.ndarray


class LTI(ABC):
    def __init__(
        self,
        nx: Annotated[int, Ge(0)],
        nu: Annotated[int, Ge(0)],
        ny: Annotated[int, Ge(0)],
        x0: Optional[npt.NDArray[np.float64]] = None,
        u0: Optional[npt.NDArray[np.float64]] = None,
    ) -> None:
        self._nx = nx
        self._nu = nu
        self._ny = ny

        self._x0 = np.zeros(nx, dtype=float) if x0 is None else x0
        self._u0 = np.zeros(nu, dtype=float) if u0 is None else u0

        if self._x0.shape[0] != nx:
            raise ValueError("x0 has invalid shape")

        if self._u0.shape[0] != nu:
            raise ValueError("u0 has invalid shape")

    @property
    def nx(self) -> int:
        return self._nx

    @property
    def nu(self) -> int:
        return self._nu

    @property
    def ny(self) -> int:
        return self._ny

    @property
    def nr(self) -> int:
        return 3 * self.nx + 2 * self.nu + self.ny

    @property
    def vr0(self) -> VR0:
        return VR0(
            x=0,
            der=self.nx,
            x0=self.nx + self.nx,
            u=self.nx + self.nx + self.nx,
            u0=self.nx + self.nx + self.nx + self.nu,
            y=self.nx + self.nx + self.nx + self.nu + self.nu,
        )

    @property
    def vr(self) -> VR:
        return VR(
            x=np.array(range(self.vr0.x, self.vr0.der), dtype=int),
            der=np.array(range(self.vr0.der, self.vr0.x0), dtype=int),
            x0=np.array(range(self.vr0.x0, self.vr0.u), dtype=int),
            u=np.array(range(self.vr0.u, self.vr0.u0), dtype=int),
            u0=np.array(range(self.vr0.u0, self.vr0.y), dtype=int),
            y=np.array(range(self.vr0.y, self.nr), dtype=int),
        )

    def has_states(self) -> bool:
        return self.nx > 0

    def has_inputs(self) -> bool:
        return self.nu > 0

    def has_outputs(self) -> bool:
        return self.ny > 0

    @property
    def x0(self) -> np.ndarray:
        return self._x0

    @property
    def u0(self) -> np.ndarray:
        return self._u0

    @abstractmethod
    def A(self) -> np.ndarray:
        raise NotImplementedError("A not implemented")

    @abstractmethod
    def B(self) -> np.ndarray:
        raise NotImplementedError("B not implemented")

    @abstractmethod
    def C(self) -> np.ndarray:
        raise NotImplementedError("C not implemented")

    @abstractmethod
    def D(self) -> np.ndarray:
        raise NotImplementedError("D not implemented")



import random

import numpy as np
import pytest

from qfmu import model


class TestStateSpace:
    nx = random.randint(1, 20)
    nu = random.randint(1, 20)
    ny = random.randint(1, 20)

    A = np.random.rand(nx, nx)
    B = np.random.rand(nx, nu)
    C = np.random.rand(ny, nx)
    D = np.random.rand(ny, nu)

    @pytest.mark.parametrize(
        "A,B,C,D,nx,nu,ny",
        [
            (A, None, None, None, nx, 0, nx),
            (None, B, None, None, nx, nu, nx),
            (None, None, None, D, 0, nu, ny),
            (A, B, None, None, nx, nu, nx),
            (A, None, C, None, nx, 0, ny),
            (A, None, None, D, nx, nu, ny),
            (None, B, C, None, nx, nu, ny),
            (None, B, None, D, nx, nu, ny),
            (None, None, C, D, nx, nu, ny),
            (A, B, C, None, nx, nu, ny),
            (A, None, C, D, nx, nu, ny),
            (None, B, C, D, nx, nu, ny),
            (A, B, C, D, nx, nu, ny),
        ],
    )
    def test_ctor(self, A, B, C, D, nx, nu, ny):
        m = model.StateSpace(A, B, C, D)
        assert m.nx == nx
        assert m.nu == nu
        assert m.ny == ny

    @pytest.mark.parametrize(
        "A,B,C,D",
        [
            (None, None, None, None),
            (None, None, C, None),
        ],
    )
    def test_invalid_ctor(self, A, B, C, D):
        with pytest.raises(ValueError):
            model.StateSpace(A, B, C, D)



import logging
from typing import Optional

import numpy as np
import numpy.typing as npt
from scipy import signal

from qfmu.model.lti import LTI


class TransferFunction(LTI):
    def __init__(
        self,
        num: npt.NDArray[np.float64],
        den: npt.NDArray[np.float64],
        x0: Optional[float] = None,
        u0: Optional[float] = None,
    ):
        self._A, self._B, self._C, self._D = signal.tf2ss(num, den)
        super().__init__(
            nx=self._A.shape[0],
            nu=1,
            ny=1,
            x0=np.array([x0]) if x0 is not None else np.zeros(self._A.shape[0]),
            u0=np.array([u0]) if u0 is not None else np.zeros(1),
        )

        logging.info(f"{signal.TransferFunction(num, den)}")

    @property
    def A(self) -> np.ndarray:
        return self._A

    @property
    def B(self) -> np.ndarray:
        return self._B

    @property
    def C(self) -> np.ndarray:
        return self._C

    @property
    def D(self) -> np.ndarray:
        return self._D



import logging
from typing import Optional

import numpy as np
import numpy.typing as npt
from scipy import signal

from qfmu.model.lti import LTI


class StateSpace(LTI):
    """Continuous-time state space system"""

    def __init__(
        self,
        A: Optional[npt.NDArray[np.float64]] = None,
        B: Optional[npt.NDArray[np.float64]] = None,
        C: Optional[npt.NDArray[np.float64]] = None,
        D: Optional[npt.NDArray[np.float64]] = None,
        x0: Optional[npt.NDArray[np.float64]] = None,
        u0: Optional[npt.NDArray[np.float64]] = None,
    ) -> None:
        """
        State space system constructor
        """
        if all([A is None, B is None, C is None, D is None]):
            raise ValueError("A, B, C, D matrices cannot all be None")
        elif all([A is not None, C is None, B is None, D is None]):
            C = np.eye(A.shape[0])
            B = np.zeros((A.shape[0], 0))
        elif all([B is not None, A is None, C is None, D is None]):
            A = np.zeros((B.shape[0], B.shape[0]))
            C = np.eye(A.shape[0])
        elif all([C is not None, A is None, B is None, D is None]):
            raise ValueError("C matrix provided without A matrix or D matrix")
        elif all([D is not None, A is None, B is None, C is None]):
            A = np.zeros((0, 0))
            C = np.zeros((D.shape[0], 0))
            B = np.zeros((0, D.shape[1]))
        elif all([A is not None, B is not None, C is None, D is None]):
            C = np.eye(A.shape[0])
        elif all([A is not None, C is not None, B is None, D is None]):
            B = np.zeros((A.shape[0], 0))

        self.m = signal.StateSpace(A, B, C, D)
        nx, nu, ny = self.m.A.shape[0], self.m.inputs, self.m.outputs
        super().__init__(nx=nx, nu=nu, ny=ny, x0=x0, u0=u0)

        logging.info(f"A[{self.A.shape[0]}, {self.A.shape[1]}] = {self.A.tolist()}")
        logging.info(f"B[{self.B.shape[0]}, {self.B.shape[1]}] = {self.B.tolist()}")
        logging.info(f"C[{self.C.shape[0]}, {self.C.shape[1]}] = {self.C.tolist()}")
        logging.info(f"D[{self.D.shape[0]}, {self.D.shape[1]}] = {self.D.tolist()}")

    @property
    def A(self) -> np.ndarray:
        return self.m.A

    @property
    def B(self) -> np.ndarray:
        return self.m.B

    @property
    def C(self) -> np.ndarray:
        return self.m.C

    @property
    def D(self) -> np.ndarray:
        return self.m.D



"""Top-level package for qFMU."""
import pathlib
import sys

__author__ = """Hang Yu"""
__email__ = "yuhang.neu@gmail.com"
__version__ = "0.2.6"

__module_path__ = pathlib.Path(__file__).parent
__template_path__ = __module_path__ / "codegen" / "templates"
__include_path__ = __module_path__ / "codegen" / "include"

if sys.platform.startswith("win"):
    __platform__ = "win"
elif sys.platform.startswith("linux"):
    __platform__ = "linux"
elif sys.platform.startswith("darwin"):
    __platform__ = "darwin"
else:
    raise Exception("Unsupported platform: " + sys.platform)

if sys.maxsize > 2**32:
    __platform__ += "64"
else:
    __platform__ += "32"



"""The setup script."""

from setuptools import find_packages, setup

install_requires = open("requirements.txt").read().strip().split("\n")
dev_requires = open("requirements_dev.txt").read().strip().split("\n")

setup(
    name="qfmu",
    description="Quickly generate an FMU from command line",
    author="Hang Yu",
    author_email="yuhang.neu@gmail.com",
    python_requires=">=3.8",
    classifiers=[
        "Natural Language :: English",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Topic :: Software Development :: Libraries",
    ],
    entry_points={
        "console_scripts": [
            "qfmu=qfmu.__main__:app",
        ],
    },
    license="BSD license",
    long_description=open("README.md", "r", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    include_package_data=True,
    keywords="qfmu",
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    package_data={"qfmu": ["codegen/templates/*", "codegen/include/*"]},
    test_suite="tests",
    install_requires=install_requires,
    extras_require={"dev": dev_requires},
    url="https://github.com/hyumo/qfmu",
    version="0.2.6",
    zip_safe=False,
)



import numpy as np


def array2cstr(arr: np.ndarray) -> str:
    return (
        np.array2string(arr, separator=",")
        .replace("[", "{")
        .replace("]", "}")
        .replace("\n", "")
    )
